/*
 * File:        Read ADC
 * Author:      Bruce Land
 * Adapted from:
 *              main.c by
 * Author:      Syed Tahmid Mahbub
 * Target PIC:  PIC32MX250F128B
 */

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config.h"
// threading library
#include "pt_cornell_1_2.h"

////////////////////////////////////
// main header file (for externs)
#include "Cardiac_neuron.h"
// graphics libraries
#include "tft_master.h"
#include "tft_gfx.h"
// need for rand function
#include <stdlib.h>
// helper function file
#include "helper_functions.h"
// main differential equations file
#include "differential_equations.h"
////////////////////////////////////


/* Demo code for interfacing TFT (ILI9340 controller) to PIC32
 * The library has been modified from a similar Adafruit library
 */
// Adafruit data:
/***************************************************
  This is an example sketch for the Adafruit 2.2" SPI display.
  This library works with the Adafruit 2.2" TFT Breakout w/SD card
  ----> http://www.adafruit.com/products/1480

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

// pullup/down macros for keypad
// PORT B
#define EnablePullUpB(bits) CNPDBCLR=bits; CNPUBSET=bits;
#define EnablePullUpA(bits) CNPDACLR=bits; CNPUASET=bits;

// Test buffers
char buffer[60];
char pressNumber[60];

enum redState {NaRed,KRed,CaRed};

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_tft, pt_button, pt_update;

static int upArrow, upPushState;
static int downArrow, downPushState;
static int rightArrow, rightPushState;
static int leftArrow, leftPushState;
static int centerButton, centerPushState;
int numberOfButtonPresses = 0;

static int screen = 0;

// First screen row logic variables
int upArrowValue = 0;
int downArrowValue = 0;
int rightArrowValue = 0;
int leftArrowValue = 0;
int centerButtonValue = 0;
int screenRow = 0;
int screenColumn = 0;

//== Timer 2 interrupt handler ===========================================
volatile unsigned int DAC_data ;// output value
int intMembraneVoltage;
volatile SpiChannel spiChn = SPI_CHANNEL2 ;	// the SPI channel to use
volatile int spiClkDiv = 2 ; // 20 MHz max speed for this DAC

volatile double membraneVoltage = -90;
volatile double totalCurrent = 0;
volatile double iNa,iCa,iK;

int sampling = 25;

// === Neuron Update Thread =============================================

static PT_THREAD (protothread_update(struct pt *pt))
{
    PT_BEGIN(pt);
      while(1) {
        // yield time 25 milliseconds
        PT_YIELD_TIME_msec(1);
        
        checkUpDownButtons(&pauseScreen,&screen);
        
        if (pauseScreen == 0){ // Oscilloscope running
            
            updatePreviousParameters();

            updateNextParameters();

            updateTimeCurrentVoltage();

            if (sampling == 0){
                plotVoltage();   
                sampling = 25;
            }

            if (scopeOutput == 0){ // display membrane voltage
                DAC_data = (int) (16.0*( (membraneVoltage) + 100.0));
            }
            else{
                if (setRed == NaRed){
                    DAC_data = (int) (16.0*( (iNa) + 100.0));
                }
                if (setRed == KRed){
                    DAC_data = (int) (16.0*( (iK) + 100.0));
                }
                if (setRed == CaRed){
                    DAC_data = (int) (16.0*( (iCa) + 100.0));
                }
            }
            // CS low to start transaction
            mPORTBClearBits(BIT_3); // start transaction

            // write to spi2
            WriteSPI2(DAC_config_chan_A | DAC_data); 

            // test for done
            while (SPI2STATbits.SPIBUSY); // wait for end of transaction

            // CS high
            mPORTBSetBits(BIT_3); // end transaction */
            
            sampling -= 1;
        }
        
        if (pauseScreen == 1){ // Get values screen
            updateFourthScreen();
            checkCursorButtons();
        }
      } // END WHILE(1) 
  PT_END(pt);
} 

// === Button Thread =============================================
// 
static PT_THREAD (protothread_button(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1){
        PT_YIELD_TIME_msec(100); // ADJUST THIS VALUE
        
        mPORTAClearBits(BIT_0|BIT_4);
        mPORTBClearBits(BIT_4|BIT_7|BIT_8); // Center, Left, Right, Up, and Down Button
        
        upArrow = mPORTBReadBits(BIT_7);
        rightArrow = mPORTAReadBits(BIT_0);  
        leftArrow = mPORTAReadBits(BIT_4);
        downArrow = mPORTBReadBits(BIT_8);
        centerButton = mPORTBReadBits(BIT_4);
        
        debounce(&downPushState,&downArrow,&downArrowValue);
        debounce(&upPushState,&upArrow,&upArrowValue);
        debounce(&leftPushState,&leftArrow,&leftArrowValue);
        debounce(&rightPushState,&rightArrow,&rightArrowValue);
        debounce(&centerPushState,&centerButton,&centerButtonValue);
        
    } //endWhile

    PT_END(pt);
}

// === TFT Thread =============================================
static PT_THREAD (protothread_tft(struct pt *pt))
{
    PT_BEGIN(pt);
    
    tft_setTextSize(1);
    
    if (screen == 0) // Concentration and conductance screen
    {  
        // yield time for adc and tft
        PT_YIELD_TIME_msec(60);
        
        clearScreenOnce();

        calculateAdc();

        updateScreenRow(&screen); // check buttons to update screen

        // Have not transitioned to next screen yet
        if (screen == 0){
            // draw black boxes for updating values
            drawBlackBoxesFirstScreen();

            updateFirstScreen(); // First screen logic based on row            
        }

    }
    
    if (screen == 1){ // Current channel screen
        PT_YIELD_TIME_msec(60);
        
        if (pauseScreen == 1){
            clearScreen = 1;
            pauseScreen = 0;
        }
        
        clearScreenOnce(); // only runs once
        
        updateScreenRow(&screen);
        
        updateSecondScreen();
        
        if (screen != 1){
            clearScreen = 1;
        }
    }
    
    if (screen == 2){ // Transition screen
        PT_YIELD_TIME_msec(60);
        
        clearScreenOnce();
        
        updateThirdScreen();
        
        updateScreenRow(&screen);
        
        if (screen != 2){
            clearScreen = 1;
        }
    }
    
    if (screen == 3){ // Error screen
        PT_YIELD_TIME_msec(60);
        
        clearScreenOnce();
        
        tft_setCursor(0,25);
        tft_writeString("Please select current channel.");
        tft_setCursor(0,50);
        tft_writeString("Press any directional arrow to return");
        
        updateScreenRow(&screen);
    }
    
  PT_END(pt);
} // animation thread

// === Main  ======================================================
void main(void) {
    SYSTEMConfigPerformance(PBCLK);
  
    ANSELA = 0; ANSELB = 0; CM1CON = 0; CM2CON = 0;

    // === config threads ==========
    // turns OFF UART support and debugger pin, unless defines are set
    PT_setup();
    
    // The PBDIV value is already set via the pragma FPBDIV option above..
    SYSTEMConfig(SYS_FREQ, SYS_CFG_WAIT_STATES | SYS_CFG_PCACHE);

    // SCK2 is pin 26 
    // SDO2 (MOSI) is in PPS output group 2, could be connected to RB5 which is pin 14
    PPSOutput(2, RPB5, SDO2);

    // control CS for DAC
    mPORTBSetPinsDigitalOut(BIT_3);
    mPORTBSetBits(BIT_3);

    // divide Fpb by 2, configure the I/O ports. Not using SS in this example
    // 16 bit transfer CKP=1 CKE=1
    // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
    // For any given peripherial, you will need to match these
    SpiChnOpen(spiChn, SPI_OPEN_ON | SPI_OPEN_MODE16 | SPI_OPEN_MSTEN | SPI_OPEN_CKE_REV , spiClkDiv);
    // === setup system wide interrupts  ========
    INTEnableSystemMultiVectoredInt();
  
    // the ADC ///////////////////////////////////////
    // configure and enable the ADC
    CloseADC10();	// ensure the ADC is off before setting the configuration

	// define setup parameters for OpenADC10
	// Turn module on | ouput in integer | trigger mode auto | enable autosample
    // ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
    // ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
    // ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
    #define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_OFF //

	// define setup parameters for OpenADC10
	// ADC ref external  | disable offset test | disable scan mode | do 1 sample | use single buf | alternate mode off
	#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF

	// Define setup parameters for OpenADC10
    // use peripherial bus clock | set sample time | set ADC clock divider
    // ADC_CONV_CLK_Tcy2 means divide CLK_PB by 2 (max speed)
    // ADC_SAMPLE_TIME_5 seems to work with a source resistance < 1kohm
    #define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_5 | ADC_CONV_CLK_Tcy2 //ADC_SAMPLE_TIME_15| ADC_CONV_CLK_Tcy2

	// define setup parameters for OpenADC10
	// set AN4 and  as analog inputs
	#define PARAM4	ENABLE_AN1_ANA // pin 26

	// define setup parameters for OpenADC10
	// do not assign channels to scan
	#define PARAM5	SKIP_SCAN_ALL

	// use ground as neg ref for A | use AN4 for input A     
	// configure to sample AN9 
	SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN1); // configure to sample AN1
	OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using the parameters defined above

	EnableADC10(); // Enable the ADC
    
    // enable RB10 as digital in for button with internal pull up
    mPORTASetPinsDigitalIn(BIT_0);
    EnablePullUpA(BIT_0);
    mPORTBSetPinsDigitalIn(BIT_4|BIT_7|BIT_8|BIT_10|BIT_13);
    EnablePullUpB(BIT_4|BIT_7|BIT_8|BIT_10|BIT_13);
   
  ///////////////////////////////////////////////////////
    
    // init the threads
    PT_INIT(&pt_tft);
    PT_INIT(&pt_button);
    PT_INIT(&pt_update);

    // init the display
    tft_init_hw();
    tft_begin();
    tft_fillScreen(ILI9340_BLACK);
    //240x320 vertical display
    tft_setRotation(1); // Use tft_setRotation(1) for 320x240

    // seed random color
    srand(1);
    
    // Initialize all differential equation parameters
    equationInit();
    
    calculateEquationConstants(); // DELETE THIS 
    
    // Initialize all three buffers with null values
    screenInit();

    // round-robin scheduler for threads
    while (1){
          if (screen == 4){
             PT_SCHEDULE(protothread_button(&pt_button));
             PT_SCHEDULE(protothread_update(&pt_update));
          }
          else {
             PT_SCHEDULE(protothread_tft(&pt_tft));
             PT_SCHEDULE(protothread_button(&pt_button));
          }
      }
} // main

// === end  ======================================================

