#include "plib.h"
#include "differential_equations.h"
#include "Cardiac_neuron.h"
#include "math.h"

//sine lookup table
volatile fix16 sin_table[sine_table_size];

// === Initialization of variables ==========================================
//All of these variables can be made adjustable to see effect on neuron
double deltaTime = .3;
int maxTime = 100;
double temperature = 37.0;

double C_M = 2.0;	//Membrane capacitance

//Intra and Extracellular Concentrations
double CAi = .0004;
double CAe = 0; // Range is 1.5-100 Typical is 2.0
double NAi = 10.0;
double NAe = 0; // Range is 10-200 Typical is 138
double Ki = 140.0;
double Ke = 0; // Range is 2.9-4.3 Typical is 4

//Base conductance (nS/pF)
double GNA = 0; // Range is 0-16 Typical is 16
double GCA = 0; // Range is 0-0.1 Typical is 0.064
double GTO = .4; 
double GK = 0; // Range is 0-1 Typical is 0.018
double GK1 = 3.9; 
double GNAB = 0.001;
double GCAB = 0.00085;
double GNAK = 1.3;
double GNACA = 1000;

//Exchange conductances
double KCa = 0.0006;
double KmCa = 1.38;
double KmKe = 1.5;
double KmNa = 87.5;
double KmNai = 10.0;
double ksat = .1;
double eta = 0.35;

//Initialize equilibrium potentials
double ENA = 0;
double EK = 0;
double ECA = 0;
double ETO = 0;

//Initialize concentration based gating constants
double fCa = 0;
double fprimeNaK = 0;
double sigmaNaK = 0;

volatile double iStim = 40.0; //inject current to elicit hodgkin huxley spikes

//Rate variables
volatile double prevVoltage;
volatile double vrate = 1;
volatile double to = 0;
volatile double f  = 1;
volatile double x = 0;
volatile double m  = 0;
volatile double prevVRATE,prevTO,prevF,prevX,prevM;

//Time-independent rate variables
volatile double k1,r,d,f_naca;

//Variables for intermediary calculations
volatile double ar,br,ad,bd,ak1,bk1,am,bm,af,bf,ato,bto;
volatile double gNa,gCa,gTO,gK,gK1;
volatile double taum,tauf,tau_to,taux,tau_vrate;
volatile double expM,expF,expVRate,expTo,expXX;
volatile double m_ss,f_ss,vrate_ss,to_ss,x_ss;
volatile double iNaK,iNaCa,iTo;

volatile double timeCounter;

// Constants used for Equilibrium Potential calculations
double R = 8.314;
double F = 96.4867;

//ALPHA_D alpha in hodgkin huxley style d gate for Ca2+ channel
double alpha_d(double voltage){
	double rate;
	double expNum;
	double denom;
		expNum = ((voltage - 22.36)/16.6813); //after getting model to work, check if this can be 16
		expNum = expNum*expNum;		
        denom = 41.8138182; //(16.6813*sqrt(2pi))
		rate = (14.9859/denom)*(exp(-.5*expNum)); //.5 is a right shift by 2
        return rate;
}

//BETA_D beta in hodgkin huxley style d gate for Ca2+ channel
double beta_d(double voltage){
	double expNum = ((voltage - 6.2744)/14.93);
	expNum = expNum*expNum;
	double denom = 37.4235; //37.4240 = 14.93*sqrt(2pi)
	double rate;
	
    rate = 0.1471 - (5.3/denom)*(exp(-.5*(expNum)));
	return rate;
}

//ALPHA_F alpha in hodgkin huxley style f gate for Ca2+ channel
double alpha_f(double voltage){
    double expVal = ((voltage - 6.1546)/6.1223);
    double rate = 0.006872/(1.0+exp(expVal));
	return rate;
}

//BETA_F beta in hodgkin huxley style f gate for Ca2+ channel
double beta_f(double v){
	double offsetV = v + 9.8255;
	double numeratorExpVal = -0.1081*offsetV;
	double denomExpVal = -0.278*offsetV;
	double exponentiationNum = exp(numeratorExpVal);
	double exponentiationDenominator = exp(denomExpVal);

	double numerator = 0.069*exponentiationNum + 0.0112;
	double denominator = 1 + exponentiationDenominator;
	double rate = ((numerator/denominator) + 0.000057474);

    return  rate;
}

double alpha_m(double voltage){
   double rate = 0; 

   double expVal = (-.1*(voltage+47.13));
   if(fabs(voltage+47.13) > 0.0001){
       rate = 0.32*(voltage+47.13) / (1 - exp(expVal));
   }
   else{
      rate = 3.2;
   }
	
   return rate;   
}

double beta_m(double voltage){
    double rate = 0.08*(exp((voltage)/(-11.0)));
	return rate;
}

//ALPHA_K1 alpha for slow potassium current gate
double alpha_k1(double v){
	double expVal =  (exp((0.06*(v - EK - 200.0))));
	double rate = 0.1/(1.0+expVal);

    return rate;
}

//BETA_K1 bK1=(3.0*exp(0.0002*(v-EK+100.0))
    //+ exp(0.1*(v-EK-10.0))) 
    // /(1.0+exp(-0.5*(v-EK))) 
double beta_k1(double v){    
    double firstExpVal = 0.0002*(v - EK + 100); //Use left shift for 2
    double secondExpVal = .1*(v - EK - 10);
    double thirdExpVal = -.5*(v-EK); //Use right shift to divide by 2
    double numerator = 3.0*exp(firstExpVal) + exp(secondExpVal);
    double denominator = 1 + exp(thirdExpVal);
    double rate = numerator/denominator;
	return rate;
}

//ALPHA_R alpha for hodgkin huxley style r gate for transient outward current
double alpha_r(double v){
	double vOffset = v-42.2912;
	double numerator = 0.5266*exp((-0.0166*vOffset));
	double denominator = 1.0 + exp((-0.0943*vOffset));
	double rate = numerator/denominator;
	return rate;
}

//BETA_R alpha for hodgkin huxley style r gate for transient outward current
double  beta_r(double v){
	double expNum = -0.1344*(v - 5.0027);
	double expDenom = (-0.1348*((v) - 5.186e-5)); //possibly approx out 5.186e-5
	double rate = ((v) * 0.00005186) + 0.5149*exp(expNum)/(1+exp(expDenom)); //possibly approximate 0.5149 as 0.5 (right shift)
	return rate;
}

//ALPHA_TO alpha for to gate of transient outward current
double alpha_to(double v){
	double vOffset = v + 34.2531;
    double expNumTop =   (-0.173*vOffset);
    double expNumBot =  (-0.1732*vOffset);
    
	double numerator = 0.0721*( exp(expNumTop))+  (0.00005612*(v));
	double denominator = 1.0+( exp(expNumBot));
	double rate = numerator/denominator;
	return rate;
}

//BETA_TO beta for to gate of transient outward current
double beta_to(double v){
	double expNum = (-0.00000000166*((v + 34.0235))); //approximately 0, exept for HUGE v
											//can approximate above out (actually have to)
	double expDenom = -0.1604*(v + 34.0235);
	double numerator = (0.0001215*v + 0.0767*(exp(expNum)));
	double denominator = (1 + (exp(expDenom)));
	double rate = numerator/denominator;
	return rate;
}

//CalcNaCaCurrent Calculate sodium calcium exchange current
double calcNaCaCurrent( double v, double T, double gNaCa, double KmNa, double Nae, double Nai, double KmCa, double ksat, double Cae, double Cai, double eta ){
	double R = 8.314; //edit out to be constants later
	double F = 96.4867;
	double RTonF = R*(T+273.15)/(F);

	double vGasRatio = v/RTonF;
	double secondTerm = exp(eta*vGasRatio);
	double thirdTerm = exp((eta-1.0)*vGasRatio);
	double totalProduct = gNaCa/( (KmNa*KmNa*KmNa+Nae*Nae*Nae)*(KmCa+Cae)*(1.0+ksat*thirdTerm));
	double INaCa1 = totalProduct*secondTerm*Nai*Nai*Nai*Cae;
	double INaCa2 = totalProduct*thirdTerm*Nae*Nae*Nae;
	double iNaCa = INaCa1 - Cai*INaCa2;
	return iNaCa;
}

//Calculate action of sodium-potassium pump
double calcNaKCurrent( double v, double sigma , double T, double gNaK ){
    double f_NaK = fNaK(v,sigma,T); 
    double iNaK = gNaK*f_NaK*.3636363636;
	return iNaK;
}

//E_TO Only use in main method; calculate equilibrium potential
double E_to( double Nae, double Ke, double Nai, double Ki,double T ){
	double E = (R*T/F) * (log((((Ke+0.043*Nae) / (Ki+0.043*Nai)))));
	return E;
}

//EQUILIBRIUMPOTENTIAL return equilibrium methods for all ions
//only run this code in the main method...no...everytime things get updated
double equilibriumPotential(double intra,double extra,double N,double T){
	double potential = R*((T)+273.15)/((N)*F) * log((extra/intra));
	return  (potential);
}

//EXPDTCHANGE We use an exponential integration method for the model. This calculates the 
//change value for the next step
double expDTChange(double DT, double tau){
	double change = exp(-DT/tau);
	return change;
}

//Calculate fCA based on permeabilities. Only calculate in main method before threads are scheduled...NO
double f_Ca(double Cai , double KCa){
	double f = KCa/(KCa + Cai);
	return f;
}

//Calculate fNaK gating variable for sodium potassium exchange
double fNaK(double v, double sigma , double T){

    double R = 8.314; //This stuff can be broken down to constants
    double F = 96.4867;
    double RTonF = R*(T+273.15)/(F);
    double vGasRatio = v/RTonF;
    double f =  (1.0/(1.0+0.1245*exp(-0.1*vGasRatio) + 0.0365*sigma*exp(-vGasRatio)));
    return f;
}

//This is a constant, calculate only at main method...NO
//Calculate f prim gating variable for sodium potassiun exchange
double f_primeNaK(double Ke, double Nai){
	double f = (1/(1 +  (pow((10/Nai),1.5)))*(Ke/(Ke+1.5)));
	return f;
}

//GATESTEADYSTATE Abstraction to maintain steady state values for ions
double gateSteadyState(double alpha,double beta){
	double ss = alpha/(alpha + beta);
	return ss;
}

//gateTau calculate time constant for gating variables
//tau = 1/(alpha + beta)
double gateTau( double alpha,double beta ){

	double tau = 1/(alpha + beta);
	return tau;
}

//getRateOfChangeK gets the change value for potassium. This follows a different pattern than all other change variables
double getRateOfChangeK( double v, double DT, double tau_x ){      
	double expRate= exp(-DT/(tau_x+40.0*(1.0-tanh(160.0+v*2.0))));
    return expRate;
}

//SIGMA_NAK intermediary constant for sodium potassium exchange. only calculate once
double sigma_nak(double Nae){
	double sigma = 0.1429*(exp(Nae/67.3)-1);
	return sigma;
}

//TAU_V Calculate time constant for gate variable v for sodium channels 
double tau_v( double v ){
	double tanhTopVal = 7.74+0.12*v;
	double tanhTop = tanh(tanhTopVal);
	double tanhBotVal = 0.07*(v+92.4);
	double tanhBot = tanh(tanhBotVal);
	double rate = .25+2.24*((1 - tanhTop)/(1-tanhBot));
	return rate;
}

//TAU_X time constant for X, rate variable for potassium
//   240*Math.exp(-((25.5+v)*(25.5+v))/156)+ ...
//   182*(1+(exp(0.154+0.0116*v)-exp(-(0.154+0.0116*v)))/ ...
//   (exp(0.154+0.0116*v)+exp(-(0.154+0.0116*v))))+ ...
//   40*(1+(exp(160+2*v)-exp(-(160+2*v)))/...
//   (exp(160+2*v)+exp(-(160+2*v))));
double tau_x( double v ){
    double tau_prime=40*(1+tanh(160+2*v));
    double expVal = -(25.5+v)*(25.5 + v)/156;
    double tanhVal = 0.154 + 0.0116*v;
    double rate = 240*exp(expVal) + 182*(1+tanh(tanhVal))+tau_prime;
	return rate;
}

//V_INF steady state value of gate variable for sodium channels 
double v_inf(double v){
	double expVal = 7.74 + 0.12*v;
	double rate = 0.5*(1-tanh(expVal)); //can use right shift for .5
	return rate;
}

//X_inf get steady steady state potassium rate variable X
double X_inf(double v){
	double rate=0.988/(1+exp(-0.861-0.0620*v));
    return rate;
}


// Called from other helper_functions.c when moving past first screen
void calculateEquationConstants(){
    ENA = equilibriumPotential(NAi,NAe,1,temperature);
	EK = equilibriumPotential(Ki,Ke,1,temperature);
	ECA = equilibriumPotential(CAi,CAe,2,temperature);
	ETO = E_to(NAe,Ke,NAi,Ki,temperature);
	fCa = f_Ca(CAi,KCa);
	fprimeNaK = f_primeNaK(Ke,NAi);
	sigmaNaK = sigma_nak(NAe);
}

// Called from main of Cardiac_neuron.c
void equationInit(){
   vrate_ss = v_inf(membraneVoltage);
   to_ss = gateSteadyState(alpha_to(membraneVoltage),beta_to(membraneVoltage));
   f_ss = gateSteadyState(alpha_f(membraneVoltage),beta_f(membraneVoltage));
   m_ss = gateSteadyState(alpha_m(membraneVoltage),beta_m(membraneVoltage));
   x_ss = X_inf(membraneVoltage);
   timeCounter = 0;
};

void updatePreviousParameters(){
    prevM = m;
    prevVRATE = vrate;
    prevTO = to;
    prevX = x;
    prevF = f;
    prevVoltage = membraneVoltage;
}

void updateNextParameters(){
    //Get R
    ar = alpha_r(membraneVoltage);
    br = beta_r(membraneVoltage);
    r = gateSteadyState(ar,br);

     //Get D
     ad = alpha_d(membraneVoltage);
     bd = beta_d(membraneVoltage);
     d = gateSteadyState(ad,bd);

     //Get K1
     ak1 = alpha_k1(membraneVoltage);
     bk1 = beta_k1(membraneVoltage);
     k1 = gateSteadyState(ak1,bk1);

    //Extract alphas and betas
    am = alpha_m(membraneVoltage);
    bm = beta_m(membraneVoltage);
    af = alpha_f(membraneVoltage);
    bf = beta_f(membraneVoltage);
    ato = alpha_to(membraneVoltage);
    bto = beta_to(membraneVoltage);

    //update conductances
    gNa = GNA * prevM * prevM * prevM * prevVRATE * prevVRATE;
    gCa =  GCA * d * prevF * fCa;
    gTO  = GTO  * r * prevTO;
    gK =  GK * prevX * prevX;
    gK1 = GK1 * k1;	//Converted to doubleing point because fixed point was failing

    //get taus
    taum = gateTau(am,bm);
    tauf = gateTau(af,bf);//gateTau(af,bf);
    tau_to = gateTau(ato,bto);
    taux = tau_x(membraneVoltage);
    tau_vrate = tau_v(membraneVoltage);

    //get change variable
    expM = expDTChange(deltaTime,taum);
    expF = expDTChange(deltaTime,tauf);
    expVRate = expDTChange(deltaTime,tau_vrate);
    expTo = expDTChange(deltaTime,tau_to);
    expXX = getRateOfChangeK(membraneVoltage,deltaTime,taux);

    //Update gate variables
    m =  m_ss + ( prevM  - m_ss ) * expM;
    f =  f_ss + ( prevF - f_ss ) * expF;
    vrate = vrate_ss + ( prevVRATE - vrate_ss ) * expVRate;
    to = to_ss + ( prevTO - to_ss ) * expTo;
    x = x_ss + ( prevX - x_ss ) * expXX;

    //update steady state values
    m_ss = gateSteadyState(am,bm);
    f_ss = gateSteadyState(af,bf);
    vrate_ss = v_inf(membraneVoltage);
    to_ss = gateSteadyState(ato,bto);
    x_ss = X_inf(membraneVoltage);

    //update currents
    iNa = (gNa + GNAB)*(membraneVoltage-ENA);
    iCa = (gCa + GCAB)* (membraneVoltage - ECA);
    iK =  (gK + gK1) * (membraneVoltage - EK);
    iNaK = calcNaKCurrent(membraneVoltage,sigmaNaK,temperature,GNAK);
    iNaCa = calcNaCaCurrent(membraneVoltage,temperature,GNACA,KmNa,NAe,NAi,KmCa,ksat,CAe,CAi,eta);
    iTo = gTO*(membraneVoltage-ETO);    
}

void updateTimeCurrentVoltage(){
    if (timeCounter >= 0.0 && timeCounter < 2.0){	
        iStim = 40;
    }
    else{
        iStim = 0;
    }
    if (timeCounter > 750.0){
        timeCounter = 0;
    }

    timeCounter = timeCounter + deltaTime;

    //Calculate total current
    totalCurrent =  iNa + iCa + iK + iNaK + iNaCa + iTo - iStim;    
     
    //Integrate to get membrane voltage
    membraneVoltage = prevVoltage - totalCurrent/C_M*deltaTime;
}